/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "Heavy_WSTD_DL3Y.h"
#include "HeavyDPF_WSTD_DL3Y.hpp"
#include <set>


#define HV_LV2_NUM_PARAMETERS 22

#define HV_HASH_NOTEIN          0x67E37CA3
#define HV_HASH_CTLIN           0x41BE0f9C
#define HV_HASH_POLYTOUCHIN     0xBC530F59
#define HV_HASH_PGMIN           0x2E1EA03D
#define HV_HASH_TOUCHIN         0x553925BD
#define HV_HASH_BENDIN          0x3083F0F7
#define HV_HASH_MIDIIN          0x149631bE
#define HV_HASH_MIDIREALTIMEIN  0x6FFF0BCF

#define HV_HASH_NOTEOUT         0xD1D4AC2
#define HV_HASH_CTLOUT          0xE5e2A040
#define HV_HASH_POLYTOUCHOUT    0xD5ACA9D1
#define HV_HASH_PGMOUT          0x8753E39E
#define HV_HASH_TOUCHOUT        0x476D4387
#define HV_HASH_BENDOUT         0xE8458013
#define HV_HASH_MIDIOUT         0x6511DE55
#define HV_HASH_MIDIOUTPORT     0x165707E4

#define MIDI_RT_CLOCK           0xF8
#define MIDI_RT_START           0xFA
#define MIDI_RT_CONTINUE        0xFB
#define MIDI_RT_STOP            0xFC
#define MIDI_RT_ACTIVESENSE     0xFE
#define MIDI_RT_RESET           0xFF

#define HV_HASH_DPF_BPM         0xDF8C2721

// midi realtime messages
std::set<int> mrtSet {
  MIDI_RT_CLOCK,
  MIDI_RT_START,
  MIDI_RT_CONTINUE,
  MIDI_RT_STOP,
  MIDI_RT_RESET
};


START_NAMESPACE_DISTRHO


// -------------------------------------------------------------------
// Heavy Send and Print hooks

static void hvSendHookFunc(HeavyContextInterface *c, const char *sendName, uint32_t sendHash, const HvMessage *m)
{
  HeavyDPF_WSTD_DL3Y* plugin = (HeavyDPF_WSTD_DL3Y*)c->getUserData();
  if (plugin != nullptr)
  {
  }
}

static void hvPrintHookFunc(HeavyContextInterface *c, const char *printLabel, const char *msgString, const HvMessage *m)
{
  char buf[64];
  char* dst = buf;
  int len = strnlen(printLabel, 48);
  dst = strncpy(dst, printLabel, len);
  dst = strcpy(dst, " ");
  dst = strncpy(dst, msgString, 63-len);
  printf("> %s \n", buf);
}

// -------------------------------------------------------------------
// Main DPF plugin class

HeavyDPF_WSTD_DL3Y::HeavyDPF_WSTD_DL3Y()
 : Plugin(HV_LV2_NUM_PARAMETERS, 0, 0)
{
  _parameters[0] = 0.0f;
  _parameters[1] = 20.0f;
  _parameters[2] = 25.0f;
  _parameters[3] = 50.0f;
  _parameters[4] = 0.0f;
  _parameters[5] = 500.0f;
  _parameters[6] = 6.0f;
  _parameters[7] = 0.0f;
  _parameters[8] = 20.0f;
  _parameters[9] = 25.0f;
  _parameters[10] = 50.0f;
  _parameters[11] = 0.0f;
  _parameters[12] = 500.0f;
  _parameters[13] = 6.0f;
  _parameters[14] = 0.0f;
  _parameters[15] = 20.0f;
  _parameters[16] = 25.0f;
  _parameters[17] = 1337.0f;
  _parameters[18] = 50.0f;
  _parameters[19] = 0.0f;
  _parameters[20] = 500.0f;
  _parameters[21] = 6.0f;
  

  _context = hv_WSTD_DL3Y_new_with_options(getSampleRate(), 10, 22, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  
  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}

HeavyDPF_WSTD_DL3Y::~HeavyDPF_WSTD_DL3Y() {
  hv_WSTD_DL3Y_free(_context);
}

void HeavyDPF_WSTD_DL3Y::initParameter(uint32_t index, Parameter& parameter)
{// initialise parameters with defaults
  switch (index)
  {
    case paramHigh:
        parameter.name = "High";
        parameter.symbol = "high";
        parameter.unit = "dB";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -15.0f;
        parameter.ranges.max = 15.0f;
        parameter.ranges.def = 0.0f;
        {
          ParameterEnumerationValue* const enumValues = new ParameterEnumerationValue[1];
          enumValues[0].value = -15.0f;
          enumValues[0].label = "-inf";
          parameter.enumValues.count = 1;
          parameter.enumValues.values = enumValues;
        }
        break;
    case paramHigh_Cross:
        parameter.name = "High Cross";
        parameter.symbol = "high_cross";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 20.0f;
        break;
    case paramHigh_Feedback:
        parameter.name = "High Feedback";
        parameter.symbol = "high_feedback";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 25.0f;
        break;
    case paramHigh_Mix:
        parameter.name = "High Mix";
        parameter.symbol = "high_mix";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 50.0f;
        break;
    case paramHigh_Sync:
        parameter.name = "High Sync";
        parameter.symbol = "high_sync";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramHigh_Time:
        parameter.name = "High Time";
        parameter.symbol = "high_time";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 50.0f;
        parameter.ranges.max = 5000.0f;
        parameter.ranges.def = 500.0f;
        break;
    case paramHigh_TimeSync:
        parameter.name = "High TimeSync";
        parameter.symbol = "high_timesync";
        parameter.hints = kParameterIsAutomatable| kParameterIsInteger;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 12.0f;
        parameter.ranges.def = 6.0f;
        
        
        if (ParameterEnumerationValue *values = new ParameterEnumerationValue[13])
        {
          parameter.enumValues.restrictedMode = true;
          values[0].value = 0.0f;
          values[0].label = "×6";
          values[1].value = 1.0f;
          values[1].label = "×5";
          values[2].value = 2.0f;
          values[2].label = "×4";
          values[3].value = 3.0f;
          values[3].label = "×3";
          values[4].value = 4.0f;
          values[4].label = "×2";
          values[5].value = 5.0f;
          values[5].label = "×1.5";
          values[6].value = 6.0f;
          values[6].label = "×1";
          values[7].value = 7.0f;
          values[7].label = "÷1.5";
          values[8].value = 8.0f;
          values[8].label = "÷2";
          values[9].value = 9.0f;
          values[9].label = "÷3";
          values[10].value = 10.0f;
          values[10].label = "÷4";
          values[11].value = 11.0f;
          values[11].label = "÷5";
          values[12].value = 12.0f;
          values[12].label = "÷6";
          parameter.enumValues.count = 13;
          parameter.enumValues.values = values;
        }
        break;
    case paramLow:
        parameter.name = "Low";
        parameter.symbol = "low";
        parameter.unit = "dB";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -15.0f;
        parameter.ranges.max = 15.0f;
        parameter.ranges.def = 0.0f;
        {
          ParameterEnumerationValue* const enumValues = new ParameterEnumerationValue[1];
          enumValues[0].value = -15.0f;
          enumValues[0].label = "-inf";
          parameter.enumValues.count = 1;
          parameter.enumValues.values = enumValues;
        }
        break;
    case paramLow_Cross:
        parameter.name = "Low Cross";
        parameter.symbol = "low_cross";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 20.0f;
        break;
    case paramLow_Feedback:
        parameter.name = "Low Feedback";
        parameter.symbol = "low_feedback";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 25.0f;
        break;
    case paramLow_Mix:
        parameter.name = "Low Mix";
        parameter.symbol = "low_mix";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 50.0f;
        break;
    case paramLow_Sync:
        parameter.name = "Low Sync";
        parameter.symbol = "low_sync";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramLow_Time:
        parameter.name = "Low Time";
        parameter.symbol = "low_time";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 50.0f;
        parameter.ranges.max = 5000.0f;
        parameter.ranges.def = 500.0f;
        break;
    case paramLow_TimeSync:
        parameter.name = "Low TimeSync";
        parameter.symbol = "low_timesync";
        parameter.hints = kParameterIsAutomatable| kParameterIsInteger;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 12.0f;
        parameter.ranges.def = 6.0f;
        
        
        if (ParameterEnumerationValue *values = new ParameterEnumerationValue[13])
        {
          parameter.enumValues.restrictedMode = true;
          values[0].value = 0.0f;
          values[0].label = "×6";
          values[1].value = 1.0f;
          values[1].label = "×5";
          values[2].value = 2.0f;
          values[2].label = "×4";
          values[3].value = 3.0f;
          values[3].label = "×3";
          values[4].value = 4.0f;
          values[4].label = "×2";
          values[5].value = 5.0f;
          values[5].label = "×1.5";
          values[6].value = 6.0f;
          values[6].label = "×1";
          values[7].value = 7.0f;
          values[7].label = "÷1.5";
          values[8].value = 8.0f;
          values[8].label = "÷2";
          values[9].value = 9.0f;
          values[9].label = "÷3";
          values[10].value = 10.0f;
          values[10].label = "÷4";
          values[11].value = 11.0f;
          values[11].label = "÷5";
          values[12].value = 12.0f;
          values[12].label = "÷6";
          parameter.enumValues.count = 13;
          parameter.enumValues.values = values;
        }
        break;
    case paramMid:
        parameter.name = "Mid";
        parameter.symbol = "mid";
        parameter.unit = "dB";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -15.0f;
        parameter.ranges.max = 15.0f;
        parameter.ranges.def = 0.0f;
        {
          ParameterEnumerationValue* const enumValues = new ParameterEnumerationValue[1];
          enumValues[0].value = -15.0f;
          enumValues[0].label = "-inf";
          parameter.enumValues.count = 1;
          parameter.enumValues.values = enumValues;
        }
        break;
    case paramMid_Cross:
        parameter.name = "Mid Cross";
        parameter.symbol = "mid_cross";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 20.0f;
        break;
    case paramMid_Feedback:
        parameter.name = "Mid Feedback";
        parameter.symbol = "mid_feedback";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 25.0f;
        break;
    case paramMid_Freq:
        parameter.name = "Mid Freq";
        parameter.symbol = "mid_freq";
        parameter.unit = "Hz";
        parameter.hints = kParameterIsAutomatable| kParameterIsLogarithmic;
        parameter.ranges.min = 313.3f;
        parameter.ranges.max = 5705.6f;
        parameter.ranges.def = 1337.0f;
        break;
    case paramMid_Mix:
        parameter.name = "Mid Mix";
        parameter.symbol = "mid_mix";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 100.0f;
        parameter.ranges.def = 50.0f;
        break;
    case paramMid_Sync:
        parameter.name = "Mid Sync";
        parameter.symbol = "mid_sync";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramMid_Time:
        parameter.name = "Mid Time";
        parameter.symbol = "mid_time";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 50.0f;
        parameter.ranges.max = 5000.0f;
        parameter.ranges.def = 500.0f;
        break;
    case paramMid_TimeSync:
        parameter.name = "Mid TimeSync";
        parameter.symbol = "mid_timesync";
        parameter.hints = kParameterIsAutomatable| kParameterIsInteger;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 12.0f;
        parameter.ranges.def = 6.0f;
        
        
        if (ParameterEnumerationValue *values = new ParameterEnumerationValue[13])
        {
          parameter.enumValues.restrictedMode = true;
          values[0].value = 0.0f;
          values[0].label = "×6";
          values[1].value = 1.0f;
          values[1].label = "×5";
          values[2].value = 2.0f;
          values[2].label = "×4";
          values[3].value = 3.0f;
          values[3].label = "×3";
          values[4].value = 4.0f;
          values[4].label = "×2";
          values[5].value = 5.0f;
          values[5].label = "×1.5";
          values[6].value = 6.0f;
          values[6].label = "×1";
          values[7].value = 7.0f;
          values[7].label = "÷1.5";
          values[8].value = 8.0f;
          values[8].label = "÷2";
          values[9].value = 9.0f;
          values[9].label = "÷3";
          values[10].value = 10.0f;
          values[10].label = "÷4";
          values[11].value = 11.0f;
          values[11].label = "÷5";
          values[12].value = 12.0f;
          values[12].label = "÷6";
          parameter.enumValues.count = 13;
          parameter.enumValues.values = values;
        }
        break;
    
  }
  
}

// -------------------------------------------------------------------
// Internal data

float HeavyDPF_WSTD_DL3Y::getParameterValue(uint32_t index) const
{
  return _parameters[index];
  
}

void HeavyDPF_WSTD_DL3Y::setParameterValue(uint32_t index, float value)
{
  switch (index) {
    case 0: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH,
          value);
      break;
    }
    case 1: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_CROSS,
          value);
      break;
    }
    case 2: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_FEEDBACK,
          value);
      break;
    }
    case 3: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_MIX,
          value);
      break;
    }
    case 4: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_SYNC,
          value);
      break;
    }
    case 5: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_TIME,
          value);
      break;
    }
    case 6: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::HIGH_TIMESYNC,
          value);
      break;
    }
    case 7: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW,
          value);
      break;
    }
    case 8: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_CROSS,
          value);
      break;
    }
    case 9: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_FEEDBACK,
          value);
      break;
    }
    case 10: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_MIX,
          value);
      break;
    }
    case 11: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_SYNC,
          value);
      break;
    }
    case 12: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_TIME,
          value);
      break;
    }
    case 13: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::LOW_TIMESYNC,
          value);
      break;
    }
    case 14: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID,
          value);
      break;
    }
    case 15: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_CROSS,
          value);
      break;
    }
    case 16: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_FEEDBACK,
          value);
      break;
    }
    case 17: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_FREQ,
          value);
      break;
    }
    case 18: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_MIX,
          value);
      break;
    }
    case 19: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_SYNC,
          value);
      break;
    }
    case 20: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_TIME,
          value);
      break;
    }
    case 21: {
      _context->sendFloatToReceiver(
          Heavy_WSTD_DL3Y::Parameter::In::MID_TIMESYNC,
          value);
      break;
    }
    
    default: return;
  }
  _parameters[index] = value;
  
}


// -------------------------------------------------------------------
// Process

// void HeavyDPF_WSTD_DL3Y::activate()
// {

// }

// void HeavyDPF_WSTD_DL3Y::deactivate()
// {

// }



// -------------------------------------------------------------------
// DPF Plugin run() loop

#if DISTRHO_PLUGIN_WANT_MIDI_INPUT
void HeavyDPF_WSTD_DL3Y::run(const float** inputs, float** outputs, uint32_t frames, const MidiEvent* midiEvents, uint32_t midiEventCount)
{
  handleMidiInput(frames, midiEvents, midiEventCount);
#else
void HeavyDPF_WSTD_DL3Y::run(const float** inputs, float** outputs, uint32_t frames)
{
#endif
  const TimePosition& timePos(getTimePosition());
  if (timePos.playing && timePos.bbt.valid)
    _context->sendMessageToReceiverV(HV_HASH_DPF_BPM, 0, "f", timePos.bbt.beatsPerMinute);

  _context->process((float**)inputs, outputs, frames);
}

// -------------------------------------------------------------------
// Callbacks

void HeavyDPF_WSTD_DL3Y::sampleRateChanged(double newSampleRate)
{
  hv_WSTD_DL3Y_free(_context);

  _context = hv_WSTD_DL3Y_new_with_options(getSampleRate(), 10, 22, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}


// -----------------------------------------------------------------------
/* Plugin entry point, called by DPF to create a new plugin instance. */

Plugin* createPlugin()
{
    return new HeavyDPF_WSTD_DL3Y();
}

// -----------------------------------------------------------------------

END_NAMESPACE_DISTRHO