/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "DistrhoUI.hpp"
#include "ResizeHandle.hpp"

START_NAMESPACE_DISTRHO

// --------------------------------------------------------------------------------------------------------------------
enum HeavyParams {
    HIGH,
    HIGH_CROSS,
    HIGH_FEEDBACK,
    HIGH_MIX,
    HIGH_SYNC,
    HIGH_TIME,
    HIGH_TIMESYNC,
    LOW,
    LOW_CROSS,
    LOW_FEEDBACK,
    LOW_MIX,
    LOW_SYNC,
    LOW_TIME,
    LOW_TIMESYNC,
    MID,
    MID_CROSS,
    MID_FEEDBACK,
    MID_FREQ,
    MID_MIX,
    MID_SYNC,
    MID_TIME,
    MID_TIMESYNC,
};

class ImGuiPluginUI : public UI
{
    
    float fhigh = 0.0f;
    float fhigh_cross = 20.0f;
    float fhigh_feedback = 25.0f;
    float fhigh_mix = 50.0f;
    bool fhigh_sync = 0.0f != 0.0f;
    float fhigh_time = 500.0f;
    int fhigh_timesync = 6.0;
    float flow = 0.0f;
    float flow_cross = 20.0f;
    float flow_feedback = 25.0f;
    float flow_mix = 50.0f;
    bool flow_sync = 0.0f != 0.0f;
    float flow_time = 500.0f;
    int flow_timesync = 6.0;
    float fmid = 0.0f;
    float fmid_cross = 20.0f;
    float fmid_feedback = 25.0f;
    float fmid_freq = 1337.0f;
    float fmid_mix = 50.0f;
    bool fmid_sync = 0.0f != 0.0f;
    float fmid_time = 500.0f;
    int fmid_timesync = 6.0;

    ResizeHandle fResizeHandle;

    // ----------------------------------------------------------------------------------------------------------------

public:
   /**
      UI class constructor.
      The UI should be initialized to a default state that matches the plugin side.
    */
    ImGuiPluginUI()
        : UI(DISTRHO_UI_DEFAULT_WIDTH, DISTRHO_UI_DEFAULT_HEIGHT),
          fResizeHandle(this)
    {
        setGeometryConstraints(DISTRHO_UI_DEFAULT_WIDTH, DISTRHO_UI_DEFAULT_HEIGHT, true);

        // hide handle if UI is resizable
        if (isResizable())
            fResizeHandle.hide();
    }

protected:
    // ----------------------------------------------------------------------------------------------------------------
    // DSP/Plugin Callbacks

   /**
      A parameter has changed on the plugin side.@n
      This is called by the host to inform the UI about parameter changes.
    */
    void parameterChanged(uint32_t index, float value) override
    {
        switch (index) {
            case HIGH:
                fhigh = value;
                break;
            case HIGH_CROSS:
                fhigh_cross = value;
                break;
            case HIGH_FEEDBACK:
                fhigh_feedback = value;
                break;
            case HIGH_MIX:
                fhigh_mix = value;
                break;
            case HIGH_SYNC:
                fhigh_sync = value != 0.0f;
                break;
            case HIGH_TIME:
                fhigh_time = value;
                break;
            case HIGH_TIMESYNC:
                fhigh_timesync = value;
                break;
            case LOW:
                flow = value;
                break;
            case LOW_CROSS:
                flow_cross = value;
                break;
            case LOW_FEEDBACK:
                flow_feedback = value;
                break;
            case LOW_MIX:
                flow_mix = value;
                break;
            case LOW_SYNC:
                flow_sync = value != 0.0f;
                break;
            case LOW_TIME:
                flow_time = value;
                break;
            case LOW_TIMESYNC:
                flow_timesync = value;
                break;
            case MID:
                fmid = value;
                break;
            case MID_CROSS:
                fmid_cross = value;
                break;
            case MID_FEEDBACK:
                fmid_feedback = value;
                break;
            case MID_FREQ:
                fmid_freq = value;
                break;
            case MID_MIX:
                fmid_mix = value;
                break;
            case MID_SYNC:
                fmid_sync = value != 0.0f;
                break;
            case MID_TIME:
                fmid_time = value;
                break;
            case MID_TIMESYNC:
                fmid_timesync = value;
                break;
            
            default: return;
        }
    
        repaint();
    }

    // ----------------------------------------------------------------------------------------------------------------
    // Widget Callbacks

   /**
      ImGui specific onDisplay function.
    */
    void onImGuiDisplay() override
    {
        const float width = getWidth();
        const float height = getHeight();
        const float margin = 20.0f * getScaleFactor();

        ImGui::SetNextWindowPos(ImVec2(margin, margin));
        ImGui::SetNextWindowSize(ImVec2(width - 2 * margin, height - 2 * margin));

        if (ImGui::Begin("WSTD DL3Y", nullptr, ImGuiWindowFlags_NoResize + ImGuiWindowFlags_NoCollapse))
        {
            if (ImGui::SliderFloat("High", &fhigh, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH, true);
                }
                setParameterValue(HIGH, fhigh);
            }
    
            if (ImGui::SliderFloat("High Cross", &fhigh_cross, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_CROSS, true);
                }
                setParameterValue(HIGH_CROSS, fhigh_cross);
            }
    
            if (ImGui::SliderFloat("High Feedback", &fhigh_feedback, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_FEEDBACK, true);
                }
                setParameterValue(HIGH_FEEDBACK, fhigh_feedback);
            }
    
            if (ImGui::SliderFloat("High Mix", &fhigh_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_MIX, true);
                }
                setParameterValue(HIGH_MIX, fhigh_mix);
            }
    
            if (ImGui::Toggle("High Sync", &fhigh_sync))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_SYNC, true);
                }
                setParameterValue(HIGH_SYNC, fhigh_sync);
            }
    
            if (ImGui::SliderFloat("High Time", &fhigh_time, 50.0f, 5000.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_TIME, true);
                }
                setParameterValue(HIGH_TIME, fhigh_time);
            }
    

            const char* high_timesync_list[13] = {
                "×6",
                "×5",
                "×4",
                "×3",
                "×2",
                "×1.5",
                "×1",
                "÷1.5",
                "÷2",
                "÷3",
                "÷4",
                "÷5",
                "÷6",
            };

            if (ImGui::BeginCombo("High TimeSync", high_timesync_list[fhigh_timesync]))
            {
                for (int n = 0; n < 13; n++)
                {
                    bool is_selected = (fhigh_timesync == n);
                    if (ImGui::Selectable(high_timesync_list[n], is_selected))
                    {
                        fhigh_timesync = n;
                        editParameter(HIGH_TIMESYNC, true);
                        setParameterValue(HIGH_TIMESYNC, fhigh_timesync);
                    }
                    if (is_selected)
                        ImGui::SetItemDefaultFocus();
                }
                ImGui::EndCombo();
            }
    
            if (ImGui::SliderFloat("Low", &flow, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW, true);
                }
                setParameterValue(LOW, flow);
            }
    
            if (ImGui::SliderFloat("Low Cross", &flow_cross, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_CROSS, true);
                }
                setParameterValue(LOW_CROSS, flow_cross);
            }
    
            if (ImGui::SliderFloat("Low Feedback", &flow_feedback, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_FEEDBACK, true);
                }
                setParameterValue(LOW_FEEDBACK, flow_feedback);
            }
    
            if (ImGui::SliderFloat("Low Mix", &flow_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_MIX, true);
                }
                setParameterValue(LOW_MIX, flow_mix);
            }
    
            if (ImGui::Toggle("Low Sync", &flow_sync))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_SYNC, true);
                }
                setParameterValue(LOW_SYNC, flow_sync);
            }
    
            if (ImGui::SliderFloat("Low Time", &flow_time, 50.0f, 5000.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_TIME, true);
                }
                setParameterValue(LOW_TIME, flow_time);
            }
    

            const char* low_timesync_list[13] = {
                "×6",
                "×5",
                "×4",
                "×3",
                "×2",
                "×1.5",
                "×1",
                "÷1.5",
                "÷2",
                "÷3",
                "÷4",
                "÷5",
                "÷6",
            };

            if (ImGui::BeginCombo("Low TimeSync", low_timesync_list[flow_timesync]))
            {
                for (int n = 0; n < 13; n++)
                {
                    bool is_selected = (flow_timesync == n);
                    if (ImGui::Selectable(low_timesync_list[n], is_selected))
                    {
                        flow_timesync = n;
                        editParameter(LOW_TIMESYNC, true);
                        setParameterValue(LOW_TIMESYNC, flow_timesync);
                    }
                    if (is_selected)
                        ImGui::SetItemDefaultFocus();
                }
                ImGui::EndCombo();
            }
    
            if (ImGui::SliderFloat("Mid", &fmid, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID, true);
                }
                setParameterValue(MID, fmid);
            }
    
            if (ImGui::SliderFloat("Mid Cross", &fmid_cross, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_CROSS, true);
                }
                setParameterValue(MID_CROSS, fmid_cross);
            }
    
            if (ImGui::SliderFloat("Mid Feedback", &fmid_feedback, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_FEEDBACK, true);
                }
                setParameterValue(MID_FEEDBACK, fmid_feedback);
            }
    
            if (ImGui::SliderFloat("Mid Freq", &fmid_freq, 313.3f, 5705.6f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_FREQ, true);
                }
                setParameterValue(MID_FREQ, fmid_freq);
            }
    
            if (ImGui::SliderFloat("Mid Mix", &fmid_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_MIX, true);
                }
                setParameterValue(MID_MIX, fmid_mix);
            }
    
            if (ImGui::Toggle("Mid Sync", &fmid_sync))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_SYNC, true);
                }
                setParameterValue(MID_SYNC, fmid_sync);
            }
    
            if (ImGui::SliderFloat("Mid Time", &fmid_time, 50.0f, 5000.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_TIME, true);
                }
                setParameterValue(MID_TIME, fmid_time);
            }
    

            const char* mid_timesync_list[13] = {
                "×6",
                "×5",
                "×4",
                "×3",
                "×2",
                "×1.5",
                "×1",
                "÷1.5",
                "÷2",
                "÷3",
                "÷4",
                "÷5",
                "÷6",
            };

            if (ImGui::BeginCombo("Mid TimeSync", mid_timesync_list[fmid_timesync]))
            {
                for (int n = 0; n < 13; n++)
                {
                    bool is_selected = (fmid_timesync == n);
                    if (ImGui::Selectable(mid_timesync_list[n], is_selected))
                    {
                        fmid_timesync = n;
                        editParameter(MID_TIMESYNC, true);
                        setParameterValue(MID_TIMESYNC, fmid_timesync);
                    }
                    if (is_selected)
                        ImGui::SetItemDefaultFocus();
                }
                ImGui::EndCombo();
            }
    
            if (ImGui::IsItemDeactivated())
            {editParameter(HIGH, false);
            editParameter(HIGH_CROSS, false);
            editParameter(HIGH_FEEDBACK, false);
            editParameter(HIGH_MIX, false);
            editParameter(HIGH_SYNC, false);
            editParameter(HIGH_TIME, false);
            editParameter(HIGH_TIMESYNC, false);
            editParameter(LOW, false);
            editParameter(LOW_CROSS, false);
            editParameter(LOW_FEEDBACK, false);
            editParameter(LOW_MIX, false);
            editParameter(LOW_SYNC, false);
            editParameter(LOW_TIME, false);
            editParameter(LOW_TIMESYNC, false);
            editParameter(MID, false);
            editParameter(MID_CROSS, false);
            editParameter(MID_FEEDBACK, false);
            editParameter(MID_FREQ, false);
            editParameter(MID_MIX, false);
            editParameter(MID_SYNC, false);
            editParameter(MID_TIME, false);
            editParameter(MID_TIMESYNC, false);
            }
        }
        ImGui::End();
    }

    DISTRHO_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(ImGuiPluginUI)
};

// --------------------------------------------------------------------------------------------------------------------

UI* createUI()
{
    return new ImGuiPluginUI();
}

// --------------------------------------------------------------------------------------------------------------------

END_NAMESPACE_DISTRHO